@@ -133,6 +133,9 @@ my $internal_id = 1;
 sub internal_id {
     my ($self, $node) = @_;
 
+    $node = $node->get(0)
+        if $node->isa('XML::LibXML::jQuery');
+
     unless ($node->hasAttribute($self->internal_id_attribute)) {
         $node->setAttribute($self->internal_id_attribute, $internal_id++);
     }
@@ -352,10 +355,10 @@ sub render  {
     # apply wrapper
     if ($self->wrapper) {
 
-        my $wrapper_document = $self->process_template($self->wrapper);
-        $wrapper_document->insert_after($element);
-        $wrapper_document->find('#content')->append($element);
-        $element = $wrapper_document;
+        my $wrapper = $self->process_template($self->wrapper);
+        $wrapper->insert_after($element);
+        $wrapper->find('#content')->append($element);
+        $element = $wrapper;
     }
 
     # rewind directive stack, then render
